# -*- coding: utf-8 -*-
#
# Copyright (C) 2020-2021 CERN.
# Copyright (C) 2020-2021 Northwestern University.
#
# Flask-Resources is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Schema that's able to handle loading data from MultiDicts."""

from marshmallow import EXCLUDE, Schema, fields, missing, pre_load
from werkzeug.datastructures import MultiDict


class MultiDictSchema(Schema):
    """MultiDict aware schema used for loading e.g. request.args."""

    class Meta:
        """Throw away unknown values."""

        unknown = EXCLUDE

    LIST_TYPES = [
        fields.List,
    ]

    @classmethod
    def is_list_field(cls, field):
        """Method used to determine if a field is a list type field.

        This is used to unpack the MultiDict into a normal dictionary.
        """
        return any((isinstance(field, t) for t in cls.LIST_TYPES))

    @pre_load
    def flatten_multidict(self, data, **kwargs):
        """Flatten a MultiDict into a normal dictionary."""
        if not isinstance(data, MultiDict):
            return data
        data = data.to_dict(flat=False)

        for name, field in self.fields.items():
            if name in data and not self.is_list_field(field):
                data[name] = data[name][0]
        return data


class BaseListSchema(Schema):
    """List Schema for dumping extra information."""

    hits = fields.Method("get_hits")
    aggregations = fields.Method("get_aggs")

    def get_hits(self, obj_list):
        """Apply hits transformation."""
        hits_list = []
        for obj in obj_list["hits"]["hits"]:
            hits_list.append(
                self.context["object_schema_cls"](context=self.context).dump(obj)
            )
        obj_list["hits"]["hits"] = hits_list
        return obj_list["hits"]

    def get_aggs(self, obj_list):
        """Apply aggregations transformation."""
        aggs = obj_list.get("aggregations")
        if not aggs:
            return missing
        return aggs


class BaseObjectSchema(Schema):
    """Base Schema for dumping extra information."""

    def dump(self, obj, **kwargs):
        """Overriding marshmallow dump to dump extra key if any."""
        object_key = self.context.get("object_key", None)
        if object_key:
            obj[object_key] = Schema.dump(self, obj, **kwargs)
            return obj
        return Schema.dump(self, obj, **kwargs)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2020-2021 CERN.
# Copyright (C) 2020-2021 Northwestern University.
#
# Flask-Resources is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Request parser for the body, headers, query string and view args."""

from .base import RequestParser
from .body import RequestBodyParser
from .decorators import request_body_parser, request_parser
from .schema import BaseListSchema, BaseObjectSchema, MultiDictSchema

__all__ = (
    "MultiDictSchema",
    "request_body_parser",
    "request_parser",
    "RequestBodyParser",
    "RequestParser",
    "BaseListSchema",
    "BaseObjectSchema",
)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2020-2021 CERN.
# Copyright (C) 2020-2021 Northwestern University.
#
# Flask-Resources is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Resources test module."""

import pytest
from flask import abort

from flask_resources import Resource, ResourceConfig, create_error_handler, route
from flask_resources.errors import HTTPJSONException


@pytest.fixture(scope="module")
def resource():
    class Config(ResourceConfig):
        blueprint_name = "hello_world"
        error_handlers = {
            403: create_error_handler(
                HTTPJSONException(code=403, description="Overwrite existing")
            ),
        }

    class HelloWorldResource(Resource):
        error_handlers = {
            Exception: create_error_handler(
                HTTPJSONException(code=500, description="Internal Server Error")
            ),
            RuntimeError: create_error_handler(
                HTTPJSONException(code=400, description="Bad request")
            ),
            403: create_error_handler(
                HTTPJSONException(code=405, description="Permission defined")
            ),
        }

        def runtime(self):
            raise RuntimeError()

        def abort403(self):
            # Used for testing overriding
            abort(403)

        def abort404(self):
            # A standard HTTPException
            abort(404)

        def unhandled(self):
            raise Exception("Boom")

        def create_url_rules(self):
            return [
                route("GET", "/runtime", self.runtime),
                route("GET", "/abort403", self.abort403),
                route("GET", "/abort404", self.abort404),
                route("GET", "/unhandled", self.unhandled),
            ]

    return HelloWorldResource(Config())


def test_error_handlers(client):
    # Test that the error was handled, and that the response is JSON
    res = client.get("/runtime")
    assert res.status_code == 400
    assert res.json["status"] == 400

    res = client.get("/abort403")
    assert res.status_code == 403
    assert res.json["status"] == 403

    res = client.get("/abort404")
    assert res.status_code == 404
    assert res.json["status"] == 404

    res = client.get("/unhandled")
    assert res.status_code == 500
    assert res.json["status"] == 500



# -*- coding: utf-8 -*-
#
# Copyright (C) 2020-2021 CERN.
# Copyright (C) 2020-2021 Northwestern University.
#
# Flask-Resources is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Serializers required interfaces."""
from copy import deepcopy


class SerializerMixin:
    """Serializer Interface."""

    def serialize_object(self, obj):
        """Serialize a single object according to the response ctx.

        The object type must implement ``SerializableMixin``.
        """
        raise NotImplementedError()

    def serialize_object_list(self, obj_list):
        """Serialize a list of objects according to the response ctx.

        Each object type of the list should implement ``SerializableMixin``.
        """
        raise NotImplementedError()


class MarshmallowSerializer(SerializerMixin):
    """Marshmallow serializer that serializes an obj into defined schema.

    :param format_serializer_cls: Serializer in charge of converting the data object into
        the desired format.
    :param object_schema_cls: Marshmallow Schema of the object.
    :param list_schema_cls: Marshmallow Schema of the object list.
    :param schema_context: Context of the Marshmallow Schema.
    """

    def __init__(
        self,
        format_serializer_cls,
        object_schema_cls,
        list_schema_cls=None,
        schema_context={},
        **serializer_options
    ):
        """Initialize the serializer."""
        self.schema_context = schema_context
        self.object_schema_cls = object_schema_cls
        self.list_schema_cls = list_schema_cls
        self.format_serializer_cls = format_serializer_cls(**serializer_options)

    def dump_obj(self, obj):
        """Dump the object using object schema class."""
        return self.object_schema_cls(context=self.schema_context).dump(obj)

    def dump_list(self, obj_list):
        """Dump the list of objects."""
        ctx = {
            "object_schema_cls": self.object_schema_cls,
        }
        ctx.update(self.schema_context)

        if self.list_schema_cls is None:
            return self.object_schema_cls(context=self.schema_context).dump(
                obj_list, many=True
            )

        return self.list_schema_cls(context=ctx).dump(obj_list)

    def serialize_object(self, obj):
        """Dump the object using the serializer."""
        return self.format_serializer_cls.serialize_object(self.dump_obj(obj))

    def serialize_object_list(self, obj_list):
        """Dump the object list using the serializer."""
        return self.format_serializer_cls.serialize_object_list(
            self.dump_list(obj_list)
        )



# -*- coding: utf-8 -*-
#
# Copyright (C) 2020 CERN.
#
# Flask--Resources is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Pytest configuration.

See https://pytest-invenio.readthedocs.io/ for documentation on which test
fixtures are available.
"""

import pytest
from flask import Flask


@pytest.fixture(scope="module")
def app(resource):
    """Doc example app."""
    app = Flask("test")
    app.register_blueprint(resource.as_blueprint())
    return app



# -*- coding: utf-8 -*-
#
# Copyright (C) 2020-2021 CERN.
# Copyright (C) 2020-2021 Northwestern University.
#
# Flask-Resources is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Request body parsers module."""

from flask import request


class RequestBodyParser:
    """Parse the request body."""

    def __init__(self, deserializer):
        """Constructor."""
        self.deserializer = deserializer

    def parse(self):
        """Parse the request body."""
        return self.deserializer.deserialize(request.data)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2020-2021 CERN.
# Copyright (C) 2020-2021 Northwestern University.
#
# Flask-Resources is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""JSON Deserializer."""

import json

from .base import DeserializerMixin


class JSONDeserializer(DeserializerMixin):
    """JSON Deserializer."""

    def deserialize(self, data):
        """Deserializes JSON into a Python dictionary."""
        return json.loads(data) if data else None



# -*- coding: utf-8 -*-
#
# Copyright (C) 2020 CERN.
# Copyright (C) 2020 Northwestern University.
#
# Flask-Resources is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Test config deferred values."""

import pytest

from flask_resources import from_conf
from flask_resources.config import resolve_from_conf


def test_from_conf():
    class Config:
        test = "val"

    assert resolve_from_conf(from_conf("test"), Config) == "val"
    assert resolve_from_conf("not-in-conf", Config) == "not-in-conf"
    pytest.raises(AttributeError, resolve_from_conf, from_conf("invalid"), Config)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2020-2023 CERN.
# Copyright (C) 2020-2021 Northwestern University.
#
# Flask-Resources is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Library for implementing configurable REST APIs.

A resource is a factory for creating Flask Blueprint that's parameterized
via a config. The main difference from a regular blueprint is:

* Syntactical overlay - it creates a slightly different way of writing
  views and wiring them up with the Flask routing system. Flask-Resources
  is meant for REST APIs and thus puts emphasis on the HTTP method, and
  as apposed to a Flask MethodView, it allows keeping all view methods
  together for all endpoints.

* Dependency injection - a resource enables easy dependency injection via
  a configuration object. The idea behind this is for instance you write
  a reusable application that you want to allow developers to customize.
  For instance you could allow a developer to accept and deserialize their
  custom XML instead of only JSON at a given endpoint while keeping the
  application view the same, or allow them to customize the URL routes via
  the Flask application config.

In addition, Flask-Resources provides basic utilities for developing REST APIs
such as:

* Content negotiation to support multiple response serializations (e.g. serving
  JSON, JSON-LD, XML from the same endpoint).

* Request parsing (query string, headers, body) using Marshmallow and data
  deserialization.

* Resource request context to enforce paradigm of passing only validated
  request data to the view function.

If you don't need any of the above, you can simply use just a normal Flask
Blueprint instead.

Below is small minimal example:

.. code-block:: python

    from flask import Flask
    from flask_resources import Resource, ResourceConfig, route

    class Config(ResourceConfig):
        blueprint_name = "hello"

    class HelloWorldResource(Resource):
        def hello_world(self):
            return "Hello, World!"

        def create_url_rules(self):
            return [
                route("GET", "/", self.hello_world),
            ]

    app = Flask('test')
    app.config.update({
        "RESOURCE_CONFIG": Config()
    })
    resource = Resource(app.config["RESOURCE_CONFIG"])
    app.register_blueprint(resource.as_blueprint())


Larger example
--------------

Below is a large example that demonstrates:

- Response handling via content negotiation.
- Error handling and mapping of business-level exceptions to JSON errors.
- Request parsing from the body content, URL query string, headers and view
  args.
- Accessing the resource request context

.. code-block:: python

    class Config(ResourceConfig):
        # Response handlers defines possible mimetypes for content
        # negotiation
        response_handlers = {
            "application/json: ResponseHandler(JSONSerializer()),
            # ...
        }

    class MyResource(Resource):

        # Error handlers maps exceptions to JSON errors.
        error_handlers = {
            ma.ValidationError: create_error_handler(
                HTTPJSONException(code=400),
            )
        }

        decorators = [
            # You can apply decorators to all views
            login_required,
        ]

        @request_parser(
            {'q': ma.fields.String(required=True)},
            # Other locations include args, view_args, headers.
            location='args',
        )
        @response_handler(many=True)
        def search(self):
            # The validated data is available in the resource request context.
            if resource_requestctx.args['q']:
                # ...
            # From the view you can return an object which the response handler
            # will serialize.
            return [], 200

        # You can parse request body depending on the Content-Type header.
        @request_body(
            parsers={
                "application/json": RequestBodyParser(JSONDeserializer())
            }
        )
        @response_handler()
        def create(self):
            return {}, 201

        # All decorators all values to come from the conf.
        @request_parser(from_conf('update_args), location='args')
        def update(self):
            return {}, 201

        def create_url_rules(self):
            return [
                route('GET', "/", self.search),
                route('POST', "/", self.create),
                route('PUT', "/<pid_value>", self.update),
                # You can selectively disable global decorators.
                route('DELETE', "/<pid_value>", self.delete, apply_decorators=False),
            ]


"""

from .config import from_conf
from .content_negotiation import with_content_negotiation
from .context import resource_requestctx
from .deserializers import JSONDeserializer
from .errors import HTTPJSONException, create_error_handler
from .parsers import (
    BaseListSchema,
    BaseObjectSchema,
    MultiDictSchema,
    RequestBodyParser,
    RequestParser,
    request_body_parser,
    request_parser,
)
from .resources import Resource, ResourceConfig, route
from .responses import ResponseHandler, response_handler
from .serializers import (
    JSONSerializer,
    MarshmallowJSONSerializer,
    MarshmallowSerializer,
)

__version__ = "0.9.0"

__all__ = (
    "__version__",
    "create_error_handler",
    "from_conf",
    "HTTPJSONException",
    "JSONDeserializer",
    "JSONSerializer",
    "MultiDictSchema",
    "request_body_parser",
    "request_parser",
    "RequestBodyParser",
    "RequestParser",
    "resource_requestctx",
    "Resource",
    "ResourceConfig",
    "response_handler",
    "ResponseHandler",
    "route",
    "with_content_negotiation",
    "BaseListSchema",
    "MarshmallowSerializer",
    "MarshmallowJSONSerializer",
    "BaseObjectSchema",
)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2020-2021 CERN.
# Copyright (C) 2020-2021 Northwestern University.
#
# Flask-Resources is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Resource request context.

The purpose of the resource request context is similar to the Flask request
context. The main difference is it serves as a state object that can hold
validated request data as well as the result of e.g. content negotiation.

The resource request context is used by default, and when it is used it
consumes all the view arguments. These can either be retrieved via a request
parser (preferably), or accessing ``request.view_args``. The goal of this
is to ensure that the view function access only validated data.
"""

from functools import wraps
from inspect import ismethod

from flask import g
from werkzeug.local import LocalProxy


#
# Proxy to the current resource context
#
def _get_context():
    """Get the resource request context from the g object."""
    if hasattr(g, "resource_requestctx"):
        return g.resource_requestctx
    raise RuntimeError("Working outside of resource request context.")


resource_requestctx = LocalProxy(_get_context)
"""Proxy to the resource's request context"""


#
# Resource context
#
class ResourceRequestCtx(object):
    """Context manager for the resource context.

    The resource request context encodes information about the currently
    executing request for a given resource, such as:

    - The mimetype selected by the content negotiation.
    - The content type of the request payload
    """

    def __init__(self, config):
        """Initialize the resource context."""
        self.config = config
        self.args = None
        self.headers = None
        self.data = None
        self.view_args = None
        self.accept_mimetype = None
        self.response_handler = None

    def __enter__(self):
        """Push the resource context manager on the current request."""
        g.resource_requestctx = self

    def __exit__(self, type, value, traceback):
        """Pop the resource context manager from the current request."""
        del g.resource_requestctx

    def update(self, values):
        """Update the context fields present in the received dictionary `values`."""
        for field, value in values.items():
            self.__setattr__(field, value)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2020-2021 CERN.
# Copyright (C) 2020-2021 Northwestern University.
#
# Flask-Resources is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Resources test module."""

import marshmallow as ma
import pytest

from flask_resources import (
    HTTPJSONException,
    Resource,
    ResourceConfig,
    create_error_handler,
    request_parser,
    resource_requestctx,
    route,
)


@pytest.fixture(scope="module")
def resource():
    class TestConfig(ResourceConfig):
        blueprint_name = "test"

    class TestResource(Resource):
        error_handlers = {
            ma.ValidationError: create_error_handler(HTTPJSONException(code=400))
        }

        @request_parser({"id": ma.fields.Int()}, location="args")
        def args(self):
            return resource_requestctx.args, 200

        @request_parser(
            {"id": ma.fields.List(ma.fields.Int())}, location="args", unknown=ma.RAISE
        )
        def multiargs(self):
            return resource_requestctx.args, 200

        @request_parser({"id": ma.fields.Int()}, location="view_args")
        def view_args(self):
            return resource_requestctx.view_args, 200

        @request_parser({"if_match": ma.fields.Int(required=True)}, location="headers")
        def header(self):
            return resource_requestctx.headers, 200

        @request_parser(
            {"if_match": ma.fields.Int(required=True)},
            location="headers",
            unknown=ma.INCLUDE,
        )
        def header_unknown(self):
            return resource_requestctx.headers, 200

        def create_url_rules(self):
            return [
                route("GET", "/args", self.args),
                route("GET", "/multiargs", self.multiargs),
                route("GET", "/viewargs/<id>", self.view_args),
                route("GET", "/header", self.header),
                route("GET", "/header-unknown", self.header_unknown),
            ]

    return TestResource(TestConfig)


def test_args_parsers(client):
    res = client.get("/args?id=1")
    assert res.json == {"id": 1}

    # Invalid value
    res = client.get("/args?id=a")
    assert res.status_code == 400

    # Unknown excluded
    res = client.get("/args?id=1&unknown=1")
    assert res.json == {"id": 1}


def test_multiargs_parsers(client):
    # Test behavior of repeated keys
    res = client.get("/args?id=1&id=2")
    assert res.json == {"id": 1}
    res = client.get("/multiargs?id=1&id=2")
    assert res.json == {"id": [1, 2]}

    # Unknown raises
    res = client.get("/multiargs?id=1&unknown=1")
    assert res.status_code == 400


def test_view_args(client):
    res = client.get("/viewargs/1")
    assert res.json == {"id": 1}
    # Invalid
    res = client.get("/viewargs/a")
    assert res.status_code == 400


def test_headers(client):
    # Test headers
    res = client.get("/header", headers={"if-match": "1"})
    assert res.json == {"if_match": 1}
    # header required
    res = client.get("/header")
    assert res.status_code == 400

    # include unknowns
    res = client.get("/header-unknown", headers={"if-match": "1"})
    assert res.json["if_match"] == 1
    assert "user_agent" in res.json



# -*- coding: utf-8 -*-
#
# Copyright (C) 2020 CERN.
# Copyright (C) 2020 Northwestern University.
#
# Flask-Resources is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Test deserialization."""

from flask import Flask
from marshmallow import Schema, fields
from speaklater import make_lazy_string

from flask_resources import (
    BaseListSchema,
    BaseObjectSchema,
    MarshmallowJSONSerializer,
    MarshmallowSerializer,
)
from flask_resources.serializers import JSONSerializer


def _(s):
    return make_lazy_string(lambda: s)


class UITestSchema(BaseObjectSchema):
    test = fields.String(dump_only=True, data_key="title_l10n")


def test_lazy_strings_are_serialized():
    serializer = JSONSerializer()
    lazy = {"key": _("Lazy")}
    list_lazy = [{"key": _("Lazy1")}, {"key": _("Lazy2")}]

    assert '{"key": "Lazy"}' == serializer.serialize_object(lazy)
    assert '[{"key": "Lazy1"}, {"key": "Lazy2"}]' == serializer.serialize_object_list(
        list_lazy
    )


def test_prettyprint():
    app = Flask("test")
    with app.test_request_context("/?prettyprint=1"):
        serializer = JSONSerializer()
        assert '{\n  "key": "1"\n}' == serializer.serialize_object({"key": "1"})


def test_marhsmallow_serializer():
    class TestSchema(Schema):
        title = fields.Str(data_key="test")

    s = MarshmallowJSONSerializer(schema_cls=TestSchema)
    assert s.serialize_object({"title": "a"}) == '{"test": "a"}'
    assert s.serialize_object_list([{"title": "a"}]) == '[{"test": "a"}]'


def test_marshmallow_serializer_without_context():
    serializer = MarshmallowSerializer(
        format_serializer_cls=JSONSerializer,
        object_schema_cls=UITestSchema,
        list_schema_cls=BaseListSchema,
    )
    test = {"test": _("test")}

    list_test = {"hits": {"hits": [{"test": _("test")}]}}
    assert serializer.serialize_object(test) == '{"title_l10n": "test"}'
    assert (
        serializer.serialize_object_list(list_test)
        == '{"hits": {"hits": [{"title_l10n": "test"}]}}'
    )


def test_marshmallow_serializer_with_context():
    serializer = MarshmallowSerializer(
        format_serializer_cls=JSONSerializer,
        object_schema_cls=UITestSchema,
        list_schema_cls=BaseListSchema,
        schema_context={"object_key": "ui"},
    )
    test = {"test": _("test")}
    list_test = {"hits": {"hits": [{"test": _("test")}]}}
    assert (
        serializer.serialize_object(test)
        == '{"test": "test", "ui": {"title_l10n": "test"}}'
    )
    assert (
        serializer.serialize_object_list(list_test)
        == '{"hits": {"hits": [{"test": "test", "ui": {"title_l10n": "test"}}]}}'
    )



# -*- coding: utf-8 -*-
#
# Copyright (C) 2020-2021 CERN.
# Copyright (C) 2020-2021 Northwestern University.
#
# Flask-Resources is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Request parser for extracting URL args, headers and view args.

The request parser uses a declarative way to extract and validate request
parameters. The parser can parse data in three different locations:

- ``args``: URL query string (i.e. ``request.args``)
- ``headers``: Request headers (i.e. ``request.headers``)
- ``view_args``: Request view args (i.e. ``request.view_args``)

The parser is not meant to parse the request body. For that you should use the
``RequestBodyParser``.

The request parser can accept both a schema or a dictionary. Using the schema
enables you to do further pre/post-processing of values, while the dict version
can be more compact.

Example with schema:

.. code-block:: python

    class MyHeaders(ma.Schema):
        content_type = ma.fields.String()

    parser = RequestParser(MyHeaders, location='headers')
    parser.parse()

Same example with dict:

.. code-block:: python

    parser = RequestParser({
        'content_type': ma.fields.String()
    }, location='headers')
    parser.parse()

**URL args parsing**

If you are parsing URL args, be aware that a query string can have repeated
variables (e.g. in ``?type=a&type=b`` the value ``type`` is repeated).

Thus if you build your own schema for URL args, you should inherit from
``MultiDictSchema``. If you don't have repeated keys you can use a normal
Marshmallow schema.

**Unknown values**

If you pass a dict for the schema, you can control what to do with unknown
values:

.. code-block:: python

    parser = RequestParser({
        'id': ma.fields.String()
    }, location='args', unknown=ma.RAISE)
    parser.parse()

If you build your own schema, the same can be achieved with by providing the
meta class:

.. code-block:: python

    class MyArgs(ma.Schema):
        id = ma.fields.String()

        class Meta:
            unknown = ma.INCLUDE
"""

import marshmallow as ma
from flask import request

from .schema import MultiDictSchema


class RequestParser:
    """Request parser."""

    def __init__(self, schema_or_dict, location, unknown=ma.EXCLUDE):
        """Constructor.

        :param schema_or_dict: A marshmallow schema class or a mapping from
            keys to fields.
        :param location: Location where to load data from. Possible values:
            (``args``, ``headers``, or ``view_args``).
        :param unknown: Determines how to handle unknown values. Possible
            values: ``ma.EXCLUDE``, ``ma.INCLUDE``, ``ma.RAISE``. Only used if
            the schema is a dict.
        """
        assert location in ["args", "headers", "view_args"]
        self._location = location
        self._unknown = unknown

        if isinstance(schema_or_dict, dict):
            self._schema = self.schema_from_dict(schema_or_dict)
        else:
            self._schema = schema_or_dict

    @property
    def location(self):
        """The request location for this request parser."""
        return self._location

    @property
    def default_schema_cls(self):
        """Get the base schema class when dynamically creating the schema.

        By default, ``request.args`` is a MultiDict which a normal Marshmallow
        schema does not know how to handle, we therefore change the schema
        only for request args parsing.

        """
        if self._location == "args":
            return MultiDictSchema
        else:
            return ma.Schema

    def schema_from_dict(self, schema_dict):
        """Construct a schema from a dict."""
        cls_ = self.default_schema_cls

        class BaseSchema(cls_):
            class Meta:
                unknown = self._unknown

        return BaseSchema.from_dict(schema_dict)

    @property
    def schema(self):
        """Build the schema class."""
        return self._schema()

    def load_data(self):
        """Load data from request."""
        if self._location == "args":
            if issubclass(self._schema, MultiDictSchema):
                return request.args
            else:
                return request.args.to_dict(flat=False)
        elif self._location == "headers":
            return {
                k.lower().replace("-", "_"): v for (k, v) in request.headers.items()
            }
        elif self._location == "view_args":
            return request.view_args
        raise RuntimeError(f"Unknown request location: {self._location}.")

    def parse(self):
        """Parse the request data."""
        return self.schema.load(self.load_data())



# -*- coding: utf-8 -*-
#
# Copyright (C) 2023 CERN.
#
# Flask-Resources is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""XML serializer."""

from .base import SerializerMixin


class XMLSerializer(SerializerMixin):
    """XML serializer implementation."""

    def __init__(self, string_encoder):
        """Initialize the JSONSerializer."""
        self._string_encoder = string_encoder

    def serialize_object(self, obj, **kwargs):
        """Dump the object into a xml string."""
        data = self.dump_obj(obj, **kwargs)
        return self.string_encoder(data)

    def serialize_object_list(self, obj_list, **kwargs):
        """Dump the object list into a xml string."""
        # this is a coupled to the `BaseListSchema` or any other list schema that
        # returns its search results with this schema
        return "\n".join(obj_list["hits"]["hits"])



# -*- coding: utf-8 -*-
#
# Copyright (C) 2020-2021 CERN.
# Copyright (C) 2020-2021 Northwestern University.
#
# Flask-Resources is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Content negotiation API."""

from functools import wraps

from flask import request
from werkzeug.datastructures import MIMEAccept

from .config import resolve_from_conf
from .context import resource_requestctx
from .errors import MIMETypeNotAccepted


class ContentNegotiator(object):
    """Content negotiation API.

    Implements a procedure for selecting a mimetype best matching what the
    client is requesting.
    """

    @classmethod
    def match(cls, mimetypes, accept_mimetypes, formats_map, fmt, default=None):
        """Select the MIME type which best matches the client request.

        :param mimetypes: Iterable of available MIME types.
        :param accept_mimetypes: The client's "Accept" header as MIMEAccept
            object.
        :param formats_map: Map of format values to MIME type.
        :param format: The client's selected format.
        :param default: Default MIMEtype if a wildcard was received.
        """
        return cls.match_by_format(formats_map, fmt) or cls.match_by_accept(
            mimetypes, accept_mimetypes, default=default
        )

    @classmethod
    def match_by_accept(cls, mimetypes, accept_mimetypes, default=None):
        """Select the MIME type which best matches Accept header.

        NOTE: Our match policy differs from Werkzeug's best_match policy:
              If the client accepts a specific mimetype and wildcards, and
              the server serves that specific mimetype, then favour
              that mimetype no matter its quality over the wildcard.
              This is as opposed to Werkzeug which only cares about quality.

        :param mimetypes: Iterable of available MIME types.
        :param accept_mimetypes: The client's "Accept" header as MIMEAccept
            object.
        :param default: Default MIMEtype if wildcard received.
        """
        assert isinstance(accept_mimetypes, MIMEAccept)
        assert "*/*" not in mimetypes

        # NOTE: accept_mimetypes is already sorted in descending quality order
        for client_mimetype, quality in accept_mimetypes:
            if client_mimetype in mimetypes:
                return client_mimetype

        # if here, then no match at all
        # WARNING: '*/*' in MIMEAccept object always evaluates to True
        # WARNING: MIMEAccept.find('*/*') always evaluates to 0
        # So we have to do the following
        accepted_values = list(accept_mimetypes.values())
        if "*/*" in accepted_values or not accepted_values:
            return default

        return None

    @classmethod
    def match_by_format(cls, formats_map, fmt):
        """Select the MIME type based on a query parameters."""
        return formats_map.get(fmt)


def with_content_negotiation(
    response_handlers=None,
    default_accept_mimetype=None,
):
    """Decorator to perform content negotiation.

    The result of the content negotiation is stored in the resources request
    context.
    """

    def decorator(f):
        @wraps(f)
        def inner_content_negotiation(*args, **kwargs):
            # Check Accept header i.e. can we even respond to the request in a common
            # mimetype?
            handlers = resolve_from_conf(response_handlers, resource_requestctx.config)
            default_mimetype = resolve_from_conf(
                default_accept_mimetype, resource_requestctx.config
            )

            accept_mimetype = ContentNegotiator.match(
                handlers.keys(),
                request.accept_mimetypes,
                {},  # TODO: Rely on config to populate this formats_map
                request.args.get("format", None),
                default_mimetype,
            )

            if not accept_mimetype:
                raise MIMETypeNotAccepted(allowed_mimetypes=handlers.keys())

            resource_requestctx.accept_mimetype = accept_mimetype
            resource_requestctx.response_handler = handlers[accept_mimetype]

            return f(*args, **kwargs)

        return inner_content_negotiation

    return decorator



# -*- coding: utf-8 -*-
#
# Copyright (C) 2020-2021 CERN.
# Copyright (C) 2020-2021 Northwestern University.
#
# Flask-Resources is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Resource view."""

from inspect import ismethod

from flask import Blueprint
from werkzeug.exceptions import HTTPException

from .config import from_conf, resolve_from_conf
from .content_negotiation import with_content_negotiation
from .context import ResourceRequestCtx
from .deserializers import JSONDeserializer
from .errors import handle_http_exception
from .parsers import RequestBodyParser
from .responses import ResponseHandler
from .serializers import JSONSerializer


def route(
    method,
    rule,
    view_meth,
    endpoint=None,
    rule_options={},
    apply_decorators=True,
):
    """Create a route.

    Use this method in ``create_url_rules()`` to build your list of rules.

    The ``view_method`` parameter should be a bound method (e.g.
    ``self.myview``).

    :param method: The HTTP method for this URL rule.
    :param rule: A URL rule.
    :param view_meth: The view method (a bound method) for this URL rule.
    :param endpoint: The name of the endpoint. By default the name is taken
        from the method name.
    :param rule_options: A dictionary of extra options passed to
        ``Blueprint.add_url_rule``.
    :param apply_decorators: Apply the decorators defined by the resource.
        Defaults to ``True``. This allows you to selective disable
        decorators which are normally applied to all view methods.
    """
    view_name = view_meth.__name__
    config = view_meth.__self__.config
    decorators = view_meth.__self__.decorators

    if apply_decorators:
        # reversed so order is the same as when applied directly to method
        for decorator in reversed(decorators):
            view_meth = decorator(view_meth)

    def view(*args, **kwargs):
        with ResourceRequestCtx(config):
            # args and kwargs are ignored on purpose - use a request parser
            # to retrieve the validated values.
            return view_meth()

    view.__name__ = view_name

    return {
        "rule": resolve_from_conf(rule, config),
        "methods": [method],
        "view_func": view,
        "endpoint": endpoint,
        **rule_options,
    }


class ResourceConfig:
    """Configuration for a resource.

    This object is used for dependency injection in a resource.
    """

    # Blueprint options and routing
    # =============================

    #: Name of the blueprint being created (used e.g. for prefix endpoint name).
    blueprint_name = None
    #: The URL prefix for the blueprint (all URL rules will be prefixed with this value)
    url_prefix = None
    #: A mapping of exception or HTTP status code to error handler functions.
    error_handlers = {}

    # Request parsing
    # ===============

    #: Request body parser (i.e. ``request.data``).
    request_body_parsers = {"application/json": RequestBodyParser(JSONDeserializer())}
    #: The default content type used to select the default request_body_parser.
    #: Set to ``None`` to require a Content-Type header.
    default_content_type = "application/json"

    # Response handling
    # =================

    #: Mapping of Accept MIME types to response handlers.
    response_handlers = {"application/json": ResponseHandler(JSONSerializer())}
    #: The default Accept MIME type if not defined by the request.
    #: Set to ``None``, to require an Accept header.
    default_accept_mimetype = "application/json"


class Resource:
    """Resource interface.

    A resource is a factory for creating Flask Blueprint that's parameterized
    via a config.
    """

    decorators = [
        with_content_negotiation(
            response_handlers=from_conf("response_handlers"),
            default_accept_mimetype=from_conf("default_accept_mimetype"),
        )
    ]
    """Decorators applied to all view functions.

    By default, the resource request context and content negotiation is
    enabled. Provide an empty list to disable them.
    """

    error_handlers = {}
    """Mapping of exceptions or HTTP codes to error handler functions.

    By default this mapping is merged with the error handlers mapping defined
    in the config.
    """

    def __init__(self, config):
        """Initialize the base resource."""
        self.config = config

    def as_blueprint(self, **options):
        """Create the blueprint with all views and error handlers.

        The method delegates to ``create_blueprint()``, ``create_url_rules()``
        and ``create_error_handlers()`` so usually you don't have to overwrite
        this method.
        """
        blueprint = self.create_blueprint(**options)

        for rule in self.create_url_rules():
            blueprint.add_url_rule(**rule)

        for exc_or_code, error_handler in self.create_error_handlers():
            blueprint.register_error_handler(exc_or_code, error_handler)

        return blueprint

    def create_blueprint(self, **options):
        """Create the blueprint.

        Override this function to customize the creation of the ``Blueprint``
        object itself.
        """
        if hasattr(self.config, "url_prefix"):
            options.setdefault("url_prefix", self.config.url_prefix)
        return Blueprint(self.config.blueprint_name, __name__, **options)

    def create_url_rules(self):
        """Create all the blueprint URL rules for this resource.

        The URL rules are registered on the blueprint using the
        ``Blueprint.add_url_rule()``.
        """
        return []

    def create_error_handlers(self):
        """Create all error handlers for this resource.

        This function should return a dictionary that maps an exception or HTTP
        response code to and error handler function. By default it merges
        error handlers defined on the resource itself with error handlers
        defined in the config. The error handlers in the config takes
        precedence over the resource defined error handlers.

        The error handlers are registered on the blueprint using the
        ``Blueprint.register_error_handler()``.
        """
        error_handlers = {
            HTTPException: handle_http_exception,
        }
        error_handlers.update(self.error_handlers)
        error_handlers.update(self.config.error_handlers)

        return error_handlers.items()



# -*- coding: utf-8 -*-
#
# Copyright (C) 2020-2021 CERN.
# Copyright (C) 2020-2021 Northwestern University.
#
# Flask-Resources is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Resources test module."""

import json

import pytest

from flask_resources import (
    JSONDeserializer,
    RequestBodyParser,
    Resource,
    ResourceConfig,
    request_body_parser,
    resource_requestctx,
    route,
)


@pytest.fixture(scope="module")
def resource():
    class TestConfig(ResourceConfig):
        blueprint_name = "test"

    class TestResource(Resource):
        @request_body_parser(
            parsers={
                "application/vnd.1": RequestBodyParser(deserializer=JSONDeserializer())
            }
        )
        def index(self):
            return resource_requestctx.data["val"], 200

        def create_url_rules(self):
            return [
                route("PUT", "/", self.index),
            ]

    return TestResource(TestConfig)


def test_body_parser(client):
    res = client.put(
        "/",
        headers={"content-type": "application/vnd.1"},
        data=json.dumps({"val": "1"}),
    )
    assert res.get_data(as_text=True) == "1"

    res = client.put(
        "/",
        headers={"content-type": "application/invalid"},
        data=json.dumps({"val": "1"}),
    )
    assert res.status_code == 415



# -*- coding: utf-8 -*-
#
# Copyright (C) 2020-2021 CERN.
# Copyright (C) 2020-2021 Northwestern University.
#
# Flask-Resources is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""JSON serializer."""

import json
import warnings

from flask import request
from flask.json import JSONEncoder as JSONEncoderBase
from speaklater import is_lazy_string

from .base import MarshmallowSerializer, SerializerMixin


def flask_request_options():
    """Options to pretty print the JSON."""
    if request and request.args.get("prettyprint"):
        return {
            "indent": 2,
            "sort_keys": True,
        }
    return {}


class JSONEncoder(JSONEncoderBase):
    """JSONEncoder for our custom needs.

    - Knows to force translate lazy translation strings.
    """

    def default(self, obj):
        """Override parent's default."""
        if is_lazy_string(obj):
            return str(obj)
        return super().default(obj)


class JSONSerializer(SerializerMixin):
    """JSON serializer implementation."""

    def __init__(self, encoder=None, options=None):
        """Initialize the JSONSerializer."""
        self._options = options or flask_request_options
        self._encoder = encoder or JSONEncoder

    @property
    def dumps_options(self):
        """Support adding options for the dumps() method."""
        return self._options() if callable(self._options) else self._options

    @property
    def encoder(self):
        """Support overriding the JSONEncoder used for serialization."""
        # We let classes through as-is
        if isinstance(self._encoder, type):
            return self._encoder
        elif callable(self._encoder):
            return self._encoder()
        return self._encoder

    def serialize_object(self, obj):
        """Dump the object into a json string."""
        return json.dumps(obj, cls=self.encoder, **self.dumps_options)

    def serialize_object_list(self, obj_list):
        """Dump the object list into a json string."""
        return json.dumps(obj_list, cls=self.encoder, **self.dumps_options)


class MarshmallowJSONSerializer(MarshmallowSerializer):
    """JSON serializing using Marshmallow to transform output."""

    warnings.warn(
        (
            "MarshmallowJSONSerializer is going to be deprecated"
            " use MarshmallowSerializer instead."
        ),
        DeprecationWarning,
    )

    def __init__(self, schema_cls, many_schema_cls=None, **options):
        """Initialize the serializer."""
        super().__init__(
            format_serializer_cls=JSONSerializer,
            object_schema_cls=schema_cls,
            list_schema_cls=many_schema_cls,
            **options
        )

    def dump_one(self, obj):
        """Dump the object with extra information."""
        warnings.warn(
            (
                "MarshmallowJSONSerializer.dump_one is going to be deprecated"
                "use MarshmallowSerializer.dump_obj instead."
            ),
            DeprecationWarning,
        )
        return super().dump_obj(obj)

    def dump_many(self, obj_list):
        """Dump the list of objects with extra information."""
        warnings.warn(
            (
                "MarshmallowJSONSerializer.dump_one is going to be deprecated"
                "use MarshmallowSerializer.dump_list instead."
            ),
            DeprecationWarning,
        )
        return super().dump_list(obj_list)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2020-2021 CERN.
# Copyright (C) 2020-2021 Northwestern University.
#
# Flask-Resources is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Response module."""

from functools import wraps

from flask import Response, make_response

from .context import resource_requestctx


def response_handler(many=False):
    """Decorator for using the response handler to create the HTTP response.

    The response handler works in conjunction with ``with_content_negotiation()``
    which is responsible for selecting the correct response handler based on
    the content negotiation.

    .. code-block:: python

        @response_handler()
        def read(self):
            return obj, 200

        @response_handler(many=True)
        def search(self):
            return [obj], 200
    """

    def decorator(f):
        @wraps(f)
        def inner(*args, **kwargs):
            res = f(*args, **kwargs)
            return resource_requestctx.response_handler.make_response(*res, many=many)

        return inner

    return decorator


class ResponseHandler:
    """Response handler which delegates to the a serializer.

    Example usage:

    .. code-block:: python

        def obj_headers(obj_or_list, code, many=False):
            return {'etag': ... }

        class Config(ResourceConfig):
            response_handlers = {
                "application/json": ResponseHandler(
                    JSONSerializer(), headers=obj_headers)
            }
    """

    def __init__(self, serializer, headers=None):
        """Constructor."""
        self.serializer = serializer
        self.headers = headers

    def make_headers(self, obj_or_list, code, many=False):
        """Builds the headers fo the response."""
        if self.headers is None:
            return {
                "content-type": resource_requestctx.accept_mimetype,
            }
        elif callable(self.headers):
            return self.headers(obj_or_list, code, many=many)
        else:
            return self.headers

    def make_response(self, obj_or_list, code, many=False):
        """Builds a response for one object."""
        # If view returns a response, bypass the serialization.
        if isinstance(obj_or_list, Response):
            return obj_or_list

        # https://flask.palletsprojects.com/en/1.1.x/api/#flask.Flask.make_response
        # (body, status, header)
        if many:
            serialize = self.serializer.serialize_object_list
        else:
            serialize = self.serializer.serialize_object

        return make_response(
            "" if obj_or_list is None else serialize(obj_or_list),
            code,
            self.make_headers(obj_or_list, code, many=many),
        )



# -*- coding: utf-8 -*-
#
# Copyright (C) 2020-2021 CERN.
# Copyright (C) 2020-2021 Northwestern University.
#
# Flask-Resources is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Serializers."""

from .base import MarshmallowSerializer, SerializerMixin
from .json import JSONSerializer, MarshmallowJSONSerializer
from .xml import XMLSerializer

__all__ = (
    "JSONSerializer",
    "MarshmallowSerializer",
    "MarshmallowJSONSerializer",
    "SerializerMixin",
    "XMLSerializer",
)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2020 CERN.
# Copyright (C) 2020 Northwestern University.
#
# Flask-Resources is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Test deserialization."""

import json

from flask_resources import JSONDeserializer


def test_json_deserializer():
    data = {"a": "test"}
    assert JSONDeserializer().deserialize(json.dumps(data)) == data

    data = {}
    assert JSONDeserializer().deserialize(json.dumps(data)) == {}

    data = None
    assert JSONDeserializer().deserialize(json.dumps(data)) is None



# -*- coding: utf-8 -*-
#
# Copyright (C) 2021 CERN.
# Copyright (C) 2021 Northwestern University.
#
# Flask-Resources is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Small utilities to resolve values from the resource configuration."""


class ConfigAttrValue:
    """Represents a value to be resolved from a config."""

    def __init__(self, config_attr):
        """Store the config attribute name."""
        self.config_attr = config_attr

    def resolve(self, config):
        """Resolve the configuration value."""
        return getattr(config, self.config_attr)


def resolve_from_conf(val, config):
    """Resolve the given value from config if needed."""
    if isinstance(val, ConfigAttrValue):
        return val.resolve(config)
    return val


def from_conf(config_attr):
    """Helper to create a config resolved value."""
    return ConfigAttrValue(config_attr)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2020-2021 CERN.
# Copyright (C) 2020-2021 Northwestern University.
#
# Flask-Resources is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Resources test module."""

import marshmallow as ma
import pytest

from flask_resources import (
    JSONSerializer,
    MarshmallowSerializer,
    Resource,
    ResourceConfig,
    ResponseHandler,
    response_handler,
    route,
)


@pytest.fixture(scope="module")
def resource():
    class TestSchema(ma.Schema):
        id = ma.fields.String()

    def my_headers(obj_or_list, code, many=False):
        return {"etag": "test", "content-type": "application/json"}

    class TestConfig(ResourceConfig):
        blueprint_name = "test"

        response_handlers = {
            "application/json": ResponseHandler(
                MarshmallowSerializer(
                    format_serializer_cls=JSONSerializer, object_schema_cls=TestSchema
                ),
                headers=my_headers,
            ),
            "application/vnd.test+json": ResponseHandler(
                MarshmallowSerializer(
                    format_serializer_cls=JSONSerializer, object_schema_cls=TestSchema
                ),
            ),
        }

    class TestResource(Resource):
        @response_handler()
        def one(self):
            return {"id": "one"}, 200

        @response_handler(many=True)
        def many(self):
            return [{"id": "many"}], 201

        def create_url_rules(self):
            return [
                route("GET", "/one", self.one),
                route("GET", "/many", self.many),
            ]

    return TestResource(TestConfig)


def test_one(client):
    res = client.get("/one", headers={"accept": "application/json"})
    assert res.headers["ETag"] == "test"
    assert res.json == {"id": "one"}

    res = client.get("/one", headers={"accept": "application/vnd.test+json"})
    assert "ETag" not in res.headers
    assert res.json == {"id": "one"}

    res = client.get("/many", headers={"accept": "application/json"})
    assert res.json == [{"id": "many"}]

    res = client.get("/one", headers={"accept": "application/xml"})
    assert res.status_code == 406



# -*- coding: utf-8 -*-
#
# Copyright (C) 2020-2021 CERN.
# Copyright (C) 2020-2021 Northwestern University.
#
# Flask-Resources is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Decorator for invoking the request parser."""

import warnings
from functools import wraps

from flask import request

from flask_resources.deserializers.json import JSONDeserializer

from ..config import resolve_from_conf
from ..context import resource_requestctx
from ..errors import InvalidContentType
from .base import RequestParser
from .body import RequestBodyParser


def request_parser(schema_or_parser, location=None, **options):
    """Create decorator for parsing the request.

    Both decorator parameters can be resolved from the resource configuration.

    :param schema_or_parser: A mapping of content types to parsers.
    :param default_content_type_name: The default content type used to select
        a parser if no content type was provided.
    """

    def decorator(f):
        @wraps(f)
        def inner(self, *args, **kwargs):
            s = resolve_from_conf(schema_or_parser, self.config)
            if isinstance(s, RequestParser):
                parser = s
                if location is not None:
                    warnings.warn("The location is ignored.")
            else:
                parser = RequestParser(s, location, **options)

            ctx_attr = getattr(resource_requestctx, parser.location)
            if ctx_attr is None:
                setattr(resource_requestctx, parser.location, parser.parse())
            else:
                ctx_attr.update(parser.parse())
            return f(self, *args, **kwargs)

        return inner

    return decorator


def request_body_parser(
    parsers={"application/json": RequestBodyParser(deserializer=JSONDeserializer())},
    default_content_type="application/json",
):
    """Create decorator for parsing the request body.

    Both decorator parameters can be resolved from the resource configuration.

    :param parsers: A mapping of content types to parsers.
    :param default_content_type_name: The default content type used to select
        a parser if no content type was provided.
    """

    def decorator(f):
        @wraps(f)
        def inner(self, *args, **kwargs):
            # Get the possible parsers
            body_parsers = resolve_from_conf(parsers, self.config)

            # Get the request body content type
            content_type = request.content_type or resolve_from_conf(
                default_content_type, self.config
            )

            # Get the parser
            parser = body_parsers.get(content_type)
            if parser is None:
                raise InvalidContentType(allowed_mimetypes=body_parsers.keys())

            # Parse the request body.
            resource_requestctx.data = parser.parse()

            return f(self, *args, **kwargs)

        return inner

    return decorator



# -*- coding: utf-8 -*-
#
# Copyright (C) 2020-2021 CERN.
# Copyright (C) 2020-2021 Northwestern University.
#
# Flask-Resources is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Resources test module."""

import pytest
from flask import Flask

from flask_resources import Resource, ResourceConfig, route


@pytest.fixture(scope="module")
def app():
    """Doc example app."""

    class Config(ResourceConfig):
        blueprint_name = "hellow"

    class HelloWorldResource(Resource):
        def hello_world(self):
            return "Hello, World!"

        def create_url_rules(self):
            return [
                route("GET", "/", self.hello_world),
            ]

    app = Flask("test")
    app.config.update({"RESOURCE_CONFIG": Config()})
    resource = HelloWorldResource(app.config["RESOURCE_CONFIG"])
    app.register_blueprint(resource.as_blueprint())
    return app


def test_doc_example(client):
    assert client.get("/").get_data(as_text=True) == "Hello, World!"
